from .common import DBObject, fetch_generator
import json
from metacat.util import epoch, validate_metadata

class DBParamCategory(DBObject):
    
    ColumnsText = "path,owner_user,owner_role,description,restricted,definitions,creator,created_timestamp"
    Columns = ColumnsText.split(",")

    """
        Definitions is JSON with the following structure:

        {
            "name": {
                "type":         'int','double','text','boolean',
                                        'int[]','double[]','text[]','boolean[]', 'dict', 'list', 'any'
                "values":       [ v1, v2, ...]      optional, ignored for boolean
                "min":          min value           optional, ignored if "values" present, ignored for boolean
                "max":          max value           optional, ignored if "values" present, ignored for boolean
            }
        }
    """

    Types =  ('int','float','text','boolean',
                'int[]','float[]','text[]','boolean[]','dict', 'list', 'any')


    def __init__(self, db, path, restricted=False, owner_role=None, owner_user=None, creator=None, definitions={}, description="", created_timestamp=None):
        self.Path = path
        self.DB = db
        self.OwnerUser = owner_user
        self.OwnerRole = owner_role
        self.Description = description
        self.Restricted = restricted
        self.Definitions = definitions         
        self.Creator = creator 
        self.CreatedTimestamp = created_timestamp
        
    def owners(self, directly=False):
        if self.OwnerUser is not None:
            return [self.OwnerUser]
        elif not directly and self.OwnerRole is not None:
            r = self.OwnerRole
            if isinstance(r, str):
                r = DBRole(self.DB, r)
            return r.members
        else:
            return []

    def to_jsonable(self):
        return dict(
            path = self.Path,
            owner_user = self.OwnerUser,
            owner_role = self.OwnerRole,
            description = self.Description,
            restricted = self.Restricted,
            definitions = self.Definitions,
            creator = self.Creator,
            created_timestamp = epoch(self.CreatedTimestamp)
        )
            
    @staticmethod
    def list(db, parent=None):
        c = db.cursor()
        columns = DBParamCategory.columns()
        if parent:
            c.execute(f"""
                select {columns}
                    from parameter_categories
                    where path like '{parent}.%'
                    order by path
            """)
        else:
            c.execute(f"""
                select {columns}
                    from parameter_categories
                    order by path
            """)
        return (DBParamCategory.from_tuple(db, tup) for tup in fetch_generator(c))

    def owned_by_user(self, user, directly=False):
        if isinstance(user, DBUser):   user = user.Username
        return user in self.owners(directly)

    def owned_by_role(self, role):
        if isinstance(role, DBRole):   role = role.name
        return self.OwnerRole == role

    def save(self, do_commit=True):
        c = self.DB.cursor()
        defs = json.dumps(self.Definitions)
        columns = self.columns()
        c.execute(f"""
            update parameter_categories
                set owner_user=%(owner_user)s, owner_role=%(owner_role)s, restricted=%(restricted)s, 
                    definitions=%(defs)s, description=%(description)s
                where path = %(path)s
            """,
            dict(path=self.Path, owner_user=self.OwnerUser, owner_role=self.OwnerRole, restricted=self.Restricted, defs=defs,
                    description=self.Description, creator=self.Creator))
        if do_commit:
            c.execute("commit")
        return self

    def create(self, do_commit=True):
        c = self.DB.cursor()
        defs = json.dumps(self.Definitions)
        columns = self.columns(exclude="created_timestamp")
        c.execute(f"""
            insert into parameter_categories({columns}) 
                values(%(path)s, %(owner_user)s, %(owner_role)s, %(description)s, %(restricted)s, %(defs)s, %(creator)s)
                returning created_timestamp
            """,
            dict(path=self.Path, owner_user=self.OwnerUser, owner_role=self.OwnerRole, restricted=self.Restricted, defs=defs,
                    description=self.Description, creator=self.Creator)
        )
        self.CreatedTimestamp = c.fetchone()[0]
        if do_commit:
            c.execute("commit")
        return self
    
    @staticmethod
    def from_tuple(db, tup):
        if tup is None: return None
        path, owner_user, owner_role, description, restricted, definitions, creator, created_timestamp = tup
        return DBParamCategory(db, path, owner_user=owner_user, owner_role=owner_role, description=description, 
                restricted=restricted, definitions=definitions, creator=creator, created_timestamp=created_timestamp)
        
    @staticmethod
    def get(db, path):
        c = db.cursor()
        columns = DBParamCategory.columns()
        c.execute(f"""
            select {columns}
                from parameter_categories where path=%s
                """, (path,)
        )
        tup = c.fetchone()
        return DBParamCategory.from_tuple(db, tup)

    @staticmethod
    def get_many(db, paths):
        c = db.cursor()
        columns = DBParamCategory.columns()
        c.execute(f"""
            select {columns}
                from parameter_categories where path=any(%s)
                """, (list(paths),)
        )
        return (DBParamCategory.from_tuple(db, tup) for tup in fetch_generator(c))

    @staticmethod
    def exists(db, path):
        return DBParamCategory.get(db, path) != None

    @staticmethod
    def category_for_path(db, path):
        # get the deepest category containing the path
        words = path.split(".")
        p = []
        paths = ['.']
        for w in words:
            if w:
                p.append(w)
                paths.append(".".join(p))
                
        c = db.cursor()
        columns = DBParamCategory.columns()
        c.execute(f"""
            select {columns}
                from parameter_categories where path = any(%s)
                order by path desc limit 1""", (paths,)
        )
        tup = c.fetchone()
        return DBParamCategory.from_tuple(db, tup)

    def validate_parameter(self, name, value):
        errors = validate_metadata(self.Definitions, self.Restricted, name=name, value=value)
        if errors:
            return False, errors[0][1]
        else:
            return True, "valid"
        return True, "valid"
    
    @staticmethod
    def validate_metadata_bulk(db, items):
        #
        # items is a list of:
        #     dictionary {"fid":}
        #     DBFile objects with metadata() method
        #     DBDataset objects with metadata() method
        #
        # returns list of tuples:
        #     (item, {"param name":"error", ...})
        #
        
        category_paths = set()
        for item in items:
            meta = item if isinstance(item, dict) else item.metadata()
            for name in meta:
                if "." in name:
                    path, _ = name.rsplit(".", 1)
                    category_paths.add(path)

        categories = {path:DBParamCategory.category_for_path(db, path) for path in category_paths}

        errors = []
        for index, item in enumerate(items):
            meta = item if isinstance(item, dict) else item.metadata()
            item_errors = []
            for name, value in meta.items():
                if "." in name:
                    path, vname = name.rsplit(".", 1)
                    category = categories[path]
                    if category is not None:
                        ok, error = category.validate_parameter(vname, value)
                        if not ok:
                            item_errors.append({"name": name, "reason": error, "value": value})
            if item_errors:
                errors.append((index, item_errors))
        return errors



from metacat.webapi import MCWebAPIError, MetaCatClient, MCError, AuthenticationError
from metacat import Version
import sys, getopt, os, json
from .cli import CLI, CLICommand

from .metacat_file import FileCLI
from .metacat_dataset import DatasetCLI
from .metacat_namespace import NamespaceCLI
from .metacat_auth import AuthCLI
from .metacat_admin import AdminCLI
from .metacat_query import QueryInterpreter
from .metacat_category import CategoryCLI
from metacat.util import validate_metadata

import warnings
warnings.simplefilter("ignore")

Usage = f"""
MetaCat version {Version}

Usage: 
    metacat [-s <server URL>] [-a <auth server URL>] command arguments
    metacat help

    Server host:port can also be specified using environment variables METACAT_SERVER_URL and METACAT_AUTH_SERVER_URL
    
    Commands are:
        auth       login, whoami, list, mydn, export, import
        dataset    create, update, list, show
        namespace  create, list, show
        file       declare, update, show, add
        query
"""

class MetaCatCLI(CLI):
    
    Opts="s:a:"    
    Usage = """[-s <server URL>] [-a <auth server URL>] <command> ...
        
            Both server and auth server URLs must be specified either using -s and -a or 
            via environment variables METACAT_SERVER_URL and METACAT_AUTH_SERVER_URL
        """
    
    def update_context(self, context, command, opts, args):

        if context is None and command != "admin":      # admin does not need to connect to the server
            server_url = opts.get("-s") or os.environ.get("METACAT_SERVER_URL")
    
            if not server_url:
                print("Server address must be specified either using -s option or using environment variable METACAT_SERVER_URL", file=sys.stderr)
                sys.exit(2)

            auth_server_url = opts.get("-a") or os.environ.get("METACAT_AUTH_SERVER_URL")
            if not auth_server_url:
                #print("Warning: MetaCat authentication server URL is not set. Using default:", server_url+"/auth", file=sys.stderr)
                auth_server_url = server_url+"/auth"
        
            context = MetaCatClient(server_url, auth_server_url)       # return the client as context
        return context

class VersionCommand(CLICommand):
    
    Usage = "print server and client versions"
    
    def __call__(self, command, client, opts, args):
        print("MetaCat Server URL:        ", client.ServerURL)
        print("Authentication server URL: ", client.AuthURL)
        print("Server version:            ", client.get_version())
        print("Client version:            ", Version)
        
class Simulate503Command(CLICommand):
    
    Usage = ""
    Hidden = True
    
    def __call__(self, command, client, opts, args):
        print(client.simulate_503())
        
class ValidateMetadataCommand(CLICommand):
    
    MinArgs = 1
    Opts = "d:"
    Usage = """[options] <JSON file with metadata>
        -d <dataset namespace>:<dataset name>           - if specified, validate the file metadata against the dataset requirements
        -q                                              - quiet - do not print anything, just use exit status to signal results
    """

    def __init__(self):
        self.Categories = None

    def find_category(self, name):
        # assume name has at least 1 dot in it
        cat_name, pname = name.rsplit('.', 1)
        cat = self.Categories.get(cat_name)
        immediate = True
        while cat_name != '.' and not cat:
            immediate = False
            words = cat_name.rsplit('.', 1)
            if len(words) == 1:
                cat_name = "."
            else:
                cat_name = words[0]
            cat = self.Categories.get(cat_name)
        return cat, immediate

    def __call__(self, command, client, opts, args):
        dataset = None
        dataset_did = opts.get("-d")
        if dataset_did:
            dataset = client.get_dataset(dataset_did)
            if dataset is None:
                print(f"Dataset {dataset_did} not found", file=sys.stderr)
                sys.exit(1)

        if self.Categories is None:
            categories = self.Categories = {c["path"]:c for c in client.list_categories()}

        meta = json.load(open(args[0], "r"))
        if not isinstance(meta, dict):
            raise InvalidArguments("Metadata must be a dictionary")
        errors = []
        for name, value in sorted(meta.items()):
            if not "." in name:
                errors.append(f"Invalid metadata parameter name: {name} - must be <category>.<name>")
                continue
            cat, immediate = self.find_category(name)
            if not cat:
                continue            # no category found, not even root
            if immediate:
                pname = name.rsplit('.', 1)[-1]
                cat_path = cat["path"]
                perrors = validate_metadata(cat["definitions"], cat["restricted"], name=pname, value=value)
                errors += [(cat_path + "." + pname, error) for (_, error) in perrors]
            elif cat["restricted"]:
                cat_path = cat["path"]
                errors.append((name, f"Undefined parameter {name} in restricted category {cat_path}"))

        if dataset is not None:
            errors += validate_metadata(dataset.get("file_meta_requirements", {}), False, meta)

        if errors:
            if "-q" not in opts:
                for name, error in errors:
                    print("%-40s: %s" % (name, error))
            sys.exit(1)
        else:
            sys.exit(0)

Commands = ["admin","auth","dataset","query","namespace","file"]

def main():

    cli = MetaCatCLI(
        "admin", AdminCLI,
        "auth", AuthCLI,
        "dataset", DatasetCLI,
        "namespace", NamespaceCLI,
        "category", CategoryCLI,
        "file", FileCLI,
        "query", QueryInterpreter,
        "version", VersionCommand(),
        "503", Simulate503Command(),
        "validate", ValidateMetadataCommand()
    )
    try:
        cli.run(sys.argv, argv0="metacat")
    except (AuthenticationError, MCError) as e:
        print(e, file=sys.stderr)
        sys.exit(1)

if __name__ == "__main__":
    main()
    



from .py3 import to_str, to_bytes
from .signed_token_jwt import SignedToken
from .token_lib import TokenLib
import time, requests, json

import urllib3      # disable "Unverified HTTPS request is being made..." warning
urllib3.disable_warnings()
del urllib3

class AuthenticationError(Exception):
    def __init__(self, message):
        self.Message = message

    def __str__(self):
        msg = "Authentication error"
        if self.Message:
            msg += ": " + self.Message
        return msg

class TokenAuthClientMixin(object):
    
    def __init__(self, service_url, auth_url, token=None, token_file=None, token_library=None):
        self.ServiceURL = service_url
        self.AuthURL = auth_url or service_url + "/auth"
        #print("TokenAuthClientMixin: AuthURL:", self.AuthURL)
        self.TokenLib = TokenLib(token_library)
        if isinstance(token, (str, bytes)):
            token = SignedToken.decode(token)

        self.TokenFile = token_file
        self.Token = token 
        if self.Token is None: self.Token = self.token()      # load from file/lib if needed
        
    def auth_headers(self):
        if self.Token:
            return {"X-Authentication-Token":self.Token.encode()}

    def token(self):
        if self.Token is None or self.Token.expiration <= time.time():
            if self.TokenFile:
                token = open(self.TokenFile, "rb").read()
                self.Token = SignedToken.decode(token)
            else:
                self.Token = self.TokenLib.get(self.ServiceURL)
        return self.Token
        
    def tokens_saved(self):
        return self.TokenLib is not None and self.TokenLib.exists()

    def login_digest(self, username, password, save_token=False):
        """Performs password-based authentication and stores the authentication token locally.
        
        Parameters
        ----------
        username : str
        password : str
            Password is not sent over the network. It is hashed and then used for digest authentication (:rfc:`2617`).

        Returns
        -------
        str
            username of the authenticated user (same as ``usernme`` argument)
        numeric
            token expiration timestamp
            
        """
        from requests.auth import HTTPDigestAuth
        auth_url = self.AuthURL
        url = "%s/%s?method=digest" % (auth_url, "auth")
        #print("login_digest: url:", url)
        response = requests.get(url, verify=False, auth=HTTPDigestAuth(username, password))
        if response.status_code != 200:
            #print(response, response.text)
            raise AuthenticationError(response.text)
        #print(response)
        #print(response.headers)
        self.Token = token = SignedToken.decode(response.headers["X-Authentication-Token"])
        #print("token:", token.Payload)
        if self.TokenLib is not None:
            self.TokenLib[self.ServerURL] = token
        return token.subject, token.expiration

    def login_ldap(self, username, password):
        """Performs password-based authentication and stores the authentication token locally using LDAP.
        
        Parameters
        ----------
        username : str
        password : str
            Password 

        Returns
        -------
        str
            username of the authenticated user (same as ``usernme`` argument)
        numeric
            token expiration timestamp
            
        """
        auth_url = self.AuthURL
        url = "%s/%s?method=ldap" % (auth_url, "auth")        
        data = b"%s:%s" % (to_bytes(username), to_bytes(password))
        #print("HTTPClient.post_json: headers:", headers)
        response = requests.post(url, verify=False, data = data)
        if response.status_code != 200:
            raise AuthenticationError(response.text)
	
        self.Token = token = SignedToken.decode(response.headers["X-Authentication-Token"])
        if self.TokenLib is not None:
            self.TokenLib[self.ServerURL] = token
        return token.subject, token.expiration

    def login_password(self, username, password):
        """Combines LDAP and :rfc:`2617` digest authentication by calling `login_ldap` first and then, if it fails, `ldap_digest` methods

        Parameters
        ----------
        username : str
        password : str
            Password 

        Returns
        -------
        str
            username of the authenticated user (same as ``usernme`` argument)
        numeric
            token expiration timestamp

        """
        user = None
        try:
            user, exp = self.login_ldap(username, password)
        except AuthenticationError:
            #print("LDAP authentication failed, trying digest")
            user, exp = self.login_digest(username, password)
        except Exception as e:
            #print(e)
            raise
        else:
            pass
            #print("Digest authentication succeeded:", user, exp)
        #print("logn_password:", user, exp)
        return user, exp
            
    def my_x509_dn(self, cert, key=None):
        auth_url = self.AuthURL
        url = f"{auth_url}/mydn"    
        cert_arg = (cert, cert) if key is None else (cert, key)
        response = requests.get(url, verify=False, cert=cert_arg)
        if response.status_code != 200:
            raise AuthenticationError(response.text)
        data = json.loads(response.text)
        return data

    def login_x509(self, username, cert, key=None):
        """Performs X.509 authentication and stores the authentication token locally.
        
        Parameters
        ----------
        username : str
        cert : str
            Path to the file with the X.509 certificate or the certificate and private key
        key : str
            Path to the file with the X.509 private key
         

        Returns
        -------
        str
            username of the authenticated user (same as ``usernme`` argument)
        numeric
            token expiration timestamp
            
        """
        auth_url = self.AuthURL
        url = f"{auth_url}/auth?method=x509&username={username}"    
        cert_arg = (cert, cert) if key is None else (cert, key)
        response = requests.get(url, verify=False, cert=cert_arg)
        if response.status_code != 200:
            raise AuthenticationError(response.text)
        self.Token = token = SignedToken.decode(response.headers["X-Authentication-Token"])
        if self.TokenLib is not None:
            self.TokenLib[self.ServerURL] = token
        return token.subject, token.expiration

    def auth_info(self):
        """Returns information about current authentication token.
        
        Returns
        -------
        str
            username of the authenticated user
        numeric
            token expiration timestamp
            
        """
        server_url = self.ServerURL
        token = self.Token
        if not token:
            if self.TokenLib:
                token = self.TokenLib.get(server_url)
        if not token:
            raise AuthenticationError("No token found")
        url = self.AuthURL + "/verify"
        response = requests.get(url, headers={
                "X-Authentication-Token":token.encode()
        })
        #print("web_api.auth_info:", response.status_code, response.text)
        if response.status_code/100 == 2:
            return token.subject, token.expiration
        else:
            raise AuthenticationError(response.text)



def undid(did, default_namespace=None):
    namespace = default_namespace
    if ':' not in did:
        name = did
    else:
        namespace, name = did.split(':', 1)
    return namespace, name

class ObjectSpec(object):

    def __init__(self, p1=None, p2=None, name=None, namespace=None, fid=None, validate=True):
        self.Params = (p1, p2, namespace, fid)
        self.DID = None
        if isinstance(p1, dict):
            if "did" in p1:
                did = self.DID = p1["did"]
                namespace, name = undid(did, namespace)
            fid = p1.get("fid")
            name = p1.get("name", name)
            namespace = p1.get("namespace", namespace)
            p1 = None
        self.FID = fid
        if p1 and p2:
            namespace, name = p1, p2
        elif p1:
            self.DID = p1
            namespace, name = undid(p1, namespace)
        self.Namespace, self.Name = namespace, name
        if validate:
            self.validate()
        
    def __str__(self):
        return f"Spec(namespace={self.Namespace}, name={self.Name}, fid={self.FID})"

    def did(self):
        if not (self.Name and self.Namespace):
            raise ValueError("The specification does not have name or namespace")
        return f"{self.Namespace}:{self.Name}"

    @staticmethod
    def from_dict(data, namespace=None, validate=True):
        name = data.get("name")
        fid = data.get("fid")
        if name:
            spec = ObjectSpec(data.get("namespace", namespace), name, fid=data.get("fid"), validate=validate)
        else:
            spec = ObjectSpec(data.get("did"), namespace=data.get("namespace", namespace), fid=data.get("fid"), validate=validate)
        return spec

    def valid(self):
        return self.Namespace and self.Name or self.FID

    def validate(self):
        if not self.valid():
            raise ValueError(f"Invalid object specification: %s, %s, namespace=%s, fid=%s" % self.Params)
        return self

    def as_dict(self):
        self.validate()
        out = {}
        if self.FID:    out["fid"] = self.FID
        if self.Name:   out["name"] = self.Name
        if self.Namespace:  out["namespace"] = self.Namespace
        return out

if __name__ == "__main__":
    for params, args in [
                (("scope", "name"), {}),
                (("scope:name",), {}),
                (("name",), {"namespace":"default"}),
                (("scope:name",), {"namespace":"default"})
            ]:
        kw = args.copy()
        print(params, kw, "->", ObjectSpec(*params, **kw))
        kw["fid"] = "abcd"
        print(params, kw, "->", ObjectSpec(*params, **kw))
    for d in [
                {   "fid": "abcd"   },
                {   "did":  "scope:name"    },
                {   "namespace":"scope", "name":"name" }
            ]:
        print(d, "->", ObjectSpec.from_dict(d))
        d["fid"] = "abcd"
        print(d, "->", ObjectSpec.from_dict(d))
        
        
        
    
    



from metacat.common.trees import Ascender, Node

class _MetaRegularizer(Ascender):
    # converts the meta expression into DNF form:
    #
    #   Node(or, [Node(and, [exp, ...])])
    #

    def _flatten_bool(self, op, nodes):
        #print("_flatten_bool: input:", nodes)
        new_nodes = []
        for c in nodes:
            #print("_flatten_bool: c:", c)
            if c.T == op:
                new_nodes += self._flatten_bool(op, c.C)
            else:
                new_nodes.append(c)
        #print("_flatten_bool: output:", new_nodes)
        return new_nodes

    def meta_or(self, node, *children):
        children = [x if x.T == "meta_and" else Node("meta_and", [x]) for x in self._flatten_bool("meta_or", children)]
        out = Node("meta_or", children)
        return out

    def _generate_and_terms(self, path, rest):
        if len(rest) == 0:  yield path
        else:
            node = rest[0]
            rest = rest[1:]
            if node.T == "meta_or":
                for c in node.C:
                    my_path = path + [c]
                    for p in self._generate_and_terms(my_path, rest):
                        yield p
            else:
                for p in self._generate_and_terms(path + [node], rest):
                    yield p

    def meta_and(self, node, *children):
        children = self._flatten_bool("meta_and", children)
        or_present = False
        for c in children:
            if c.T == "meta_or":
                or_present = True
                break

        if or_present:
            paths = list(self._generate_and_terms([], children))
            #print("paths:")
            #for p in paths:
            #    print(p)
            paths = [self._flatten_bool("meta_and", p) for p in paths]
            #print("meta_and: children:", paths)
            return Node("meta_or", [Node("meta_and", p) for p in paths])
        else:
            return Node("meta_and", children)

    def __default__(self, typ, children, meta):
        return Node("meta_or", [
            Node("meta_and", [Node(typ, children, _meta=meta)])
        ])

class MetaExpressionDNF(object):
    
    def __init__(self, exp):
        #
        # meta_exp is a nested list representing the query filter expression in DNF:
        #
        # meta_exp = [meta_or, ...]
        # meta_or = [meta_and, ...]
        # meta_and = [(op, aname, avalue), ...]
        #
        self.Exp = None
        self.DNF = None
        if exp is not None:
            #
            # converts canonic Node expression (meta_or of one or more meta_ands) into nested or-list or and-lists
            #
            #assert isinstance(self.Exp, Node)
            assert exp.T in ("meta_or", "meta_and")
            exp = self.regularize(exp)
            assert exp.T == "meta_or" and all (c.T == "meta_and" for c in exp.C or [])
            self.DNF = [and_item.C for and_item in exp.C]

        #print("MetaExpressionDNF: exp:", self.DNF)
        #self.validate_exp(meta_exp)
        
    def __str__(self):
        return self.file_ids_sql()
        
    __repr__= __str__
    
    @staticmethod
    def regularize(exp):
        return _MetaRegularizer()(exp)

    def sql_and(self, and_terms, table_name, meta_column_name="metadata"):

        def sql_literal(v):
            if isinstance(v, str):       
                v = "'%s'" % (v,)
            elif isinstance(v, bool):    v = "true" if v else "false"
            elif v is None:              v = "null"
            else:   v = str(v)
            return v
            
        def json_literal(v):
            if isinstance(v, str):       v = '"%s"' % (v,)
            else:   v = sql_literal(v)
            return v
            
        def pg_type(v):
            if isinstance(v, bool):   pgtype='boolean'
            elif isinstance(v, str):   pgtype='text'
            elif isinstance(v, int):   pgtype='bigint'
            elif isinstance(v, float):   pgtype='double precision'
            else:
                raise ValueError("Unrecognized literal type: %s %s" % (v, type(v)))
            return pgtype
            
        contains_items = []
        parts = []
        
        for exp in and_terms:
            
            op = exp.T
            args = exp.C
            negate = False

            term = ""

            if op == "present":
                aname = exp["name"]
                if not '.' in aname:
                    term = "true" if aname in DBFile.ColumnAttributes else "false"
                else:
                    term = f"{table_name}.{meta_column_name} ? '{aname}'"

            elif op == "not_present":
                aname = exp["name"]
                if not '.' in aname:
                    term = "false" if aname in DBFile.ColumnAttributes else "true"
                else:
                    term = f"{table_name}.{meta_column_name} ? '{aname}'"
            
            else:
                assert op in ("cmp_op", "in_range", "in_set", "not_in_range", "not_in_set"), f"Unexpected expression type: {op}, exp:\n" + exp.pretty()
                arg = args[0]
                assert arg.T in ("array_any", "array_subscript", "array_length", "object_attribute", "meta_attribute")

                negate = not not exp.get("neg")
                aname = arg["name"]
                
                if arg.T == "array_subscript":
                    # a[i] = x
                    aname, inx = arg["name"], arg["index"]
                    inx = json_literal(inx)
                    subscript = f"[{inx}]"
                elif arg.T == "array_any":
                    aname = arg["name"]
                    subscript = "[*]"
                elif arg.T in ("meta_attribute", "object_attribute"):
                    aname = arg["name"]
                    subscript = ""
                elif arg.T == "array_length":
                    aname = arg["name"]
                else:
                    raise ValueError(f"Unrecognozed argument type \"{arg.T}\"")

                # - query time slows down significantly if this is addded
                #if arg.T in ("array_subscript", "array_any", "array_all"):
                #    # require that "aname" is an array, not just a scalar
                #    parts.append(f"{table_name}.{meta_column_name} @> '{{\"{aname}\":[]}}'")
                
                if op == "in_range":
                    assert len(args) == 1
                    typ, low, high = exp["type"], exp["low"], exp["high"]
                    if typ == "date_constant":
                        high = float(high + 3600*24 - 0.0001)
                    if arg.T == "object_attribute":
                        low = sql_literal(low)
                        high = sql_literal(high)
                    else:
                        low = json_literal(low)
                        high = json_literal(high)
                    if arg.T == "object_attribute":
                        term = f"{table_name}.{aname} between {low} and {high}"
                    elif arg.T in ("array_subscript", "scalar", "array_any"):
                        term = f"{table_name}.{meta_column_name} @? '$.\"{aname}\"{subscript} ? (@ >= {low} && @ <= {high})'"
                    elif arg.T == "array_length":
                        n = "not" if negate else ""
                        negate = False
                        term = f"jsonb_array_length({table_name}.{meta_column_name} -> '{aname}') {n} between {low} and {high}"
                        
                if op == "not_in_range":
                    assert len(args) == 1
                    typ, low, high = exp["type"], exp["low"], exp["high"]
                    if typ == "date_constant":
                        high = float(high + 3600*24 - 0.0001)
                    if arg.T == "object_attribute":
                        low = sql_literal(low)
                        high = sql_literal(high)
                    else:
                        low = json_literal(low)
                        high = json_literal(high)
                    if arg.T == "object_attribute":
                        term = f"not ({table_name}.{aname} between {low} and {high})"
                    elif arg.T in ("array_subscript", "scalar", "array_any"):
                        term = f"{table_name}.{meta_column_name} @? '$.\"{aname}\"{subscript} ? (@ < {low} || @ > {high})'"
                    elif arg.T == "array_length":
                        n = "" if negate else "not"
                        negate = False
                        term = f"jsonb_array_length({table_name}.{meta_column_name} -> '{aname}') {n} between {low} and {high}"
                        
                elif op == "in_set":
                    if arg.T == "object_attribute":
                        values = [sql_literal(v) for v in exp["set"]]
                        value_list = ",".join(values)
                        term = f"{table_name}.{aname} in ({value_list})"
                    elif arg.T == "array_length":
                        values = [sql_literal(v) for v in exp["set"]]
                        value_list = ",".join(values)
                        n = "not" if negate else ""
                        negate = False
                        term = f"jsonb_array_length({table_name}.{meta_column_name} -> '{aname}') {n} in ({value_list})"
                    else:           # arg.T in ("array_any", "array_subscript","scalar")
                        values = [json_literal(x) for x in exp["set"]]
                        or_parts = [f"@ == {v}" for v in values]
                        predicate = " || ".join(or_parts)
                        term = f"{table_name}.{meta_column_name} @? '$.\"{aname}\"{subscript} ? ({predicate})'"
                        
                elif op == "not_in_set":
                    if arg.T == "object_attribute":
                        values = [sql_literal(v) for v in exp["set"]]
                        value_list = ",".join(values)
                        term = f"not ({table_name}.{aname} in ({value_list}))"
                    elif arg.T == "array_length":
                        values = [sql_literal(v) for v in exp["set"]]
                        value_list = ",".join(values)
                        n = "" if negate else "not"
                        negate = False
                        term = f"not(jsonb_array_length({table_name}.{meta_column_name} -> '{aname}') {n} in ({value_list}))"
                    else:           # arg.T in ("array_any", "array_subscript","scalar")
                        values = [json_literal(x) for x in exp["set"]]
                        and_parts = [f"@ != {v}" for v in values]
                        predicate = " && ".join(and_parts)
                        term = f"{table_name}.{meta_column_name} @? '$.\"{aname}\"{subscript} ? ({predicate})'"
                        
                elif op == "cmp_op":
                    cmp_op = exp["op"]
                    if cmp_op == '=': cmp_op = "=="
                    sql_cmp_op = "=" if cmp_op == "==" else cmp_op
                    value = args[1]
                    value_type, value = value.T, value["value"]
                    sql_value = sql_literal(value)
                    value = json_literal(value)
                    
                    if arg.T == "object_attribute":
                        term = f"{table_name}.{aname} {sql_cmp_op} {sql_value}"
                    elif arg.T == "array_length":
                        term = f"jsonb_array_length({table_name}.{meta_column_name} -> '{aname}') {sql_cmp_op} {value}"
                    elif value_type == "date_constant":
                        assert cmp_op in ("<", "<=", ">", ">=", "=", "==", "!=")
                        if cmp_op in ("=", "=="):
                            v1 = value + 3600*24
                            term = f"{table_name}.{meta_column_name} @? '$.\"{aname}\"{subscript} ? (@ < {v1} && @ >= {value})'"
                        elif cmp_op == "!=":
                            v1 = value + 3600*24
                            term = f"{table_name}.{meta_column_name} @? '$.\"{aname}\"{subscript} ? (@ >= {v1} || @ < {value})'"
                        else:
                            if cmp_op == ">":
                                value += 3600*24
                                cmp_op = ">="
                            elif cmp_op == "<=":
                                value += 3600*24
                                cmp_op = "<"
                            term = f"{table_name}.{meta_column_name} @@ '$.\"{aname}\"{subscript} {cmp_op} {value}'"
                    elif cmp_op in ("~", "~*", "!~", "!~*"):
                        negate_predicate = False
                        if cmp_op.startswith('!'):
                            cmp_op = cmp_op[1:]
                            negate_predicate = not negate_predicate
                        flags = ' flag "i"' if cmp_op.endswith("*") else ''
                        cmp_op = "like_regex"
                        value = f"{value}{flags}"
                    
                        predicate = f"@ like_regex {value} {flags}"
                        if negate_predicate: 
                            predicate = f"!({predicate})"
                        term = f"{table_name}.{meta_column_name} @? '$.\"{aname}\"{subscript} ? ({predicate})'"
                    else:
                        # scalar, array_subscript, array_any
                        term = f"{table_name}.{meta_column_name} @@ '$.\"{aname}\"{subscript} {cmp_op} {value}'"
                    
            if negate:  term = f"not ({term})"
            parts.append(term)

        if contains_items:
            parts.append("%s.{meta_column_name} @> '{%s}'" % (table_name, ",".join(contains_items )))
        return parts

    def sql(self, table_name, meta_column_name="metadata"):
        if not self.DNF:
            return None
        else:
            out = []
            for i, or_part in enumerate(self.DNF):
                and_parts = self.sql_and(or_part, table_name, meta_column_name)
                for j, and_part in enumerate(and_parts):
                    #print("and_part:", and_part)
                    if i == 0:
                        prefix = "" if j == 0 else "and "
                    else:
                        prefix = "or  " if j == 0 else "    and "
                    out.append(f"{prefix}( {and_part} )")
            out = "\n".join(out)
            #print("returning:\n", out)
            return out
            



from textwrap import dedent, indent

def chunked(iterable, n):
    if isinstance(iterable, (list, tuple)):
        for i in range(0, len(iterable), n):
            yield iterable[i:i+n]
    else:
        chunk = []
        for item in iterable:
            chunk.append(item)
            if len(chunk) >= n:
                yield chunk
                chunk = []
        if chunk:
            yield chunk
            
def unique(iterable, key=None):
    seen = set()
    for item in iterable:
        k = item if key is None else key(item)
        if not k in seen:
            seen.add(k)
            yield item

def first_not_empty(lst):
    val = None
    for v in lst:
        val = v
        if v is not None and not (isinstance(v, list) and len(v) == 0):
            return v
    else:
        return val
        
def limited(iterable, n):
    if n is None:
        yield from iterable
    if isinstance(iterable, (list, tuple)):
        yield from iterable[:n]
    for f in iterable:
        if n is None:
            yield f
        else:
            if n or n > 0:
                yield f
            else:
                break
            n -= 1
            
def strided(iterable, n, i=0):
    if n is None:
        yield from iterable
    for j, f in enumerate(iterable):
        if j%n == i:
            yield f
            
def skipped(iterable, n):
    if n is None:
        yield from iterable
    if isinstance(iterable, (list, tuple)):
        yield from iterable[n:]
    for f in iterable:
        if n > 0:
            n -= 1
        else:
            yield f

def insert_sql_one(outer, tag, text):
    lines = dedent(outer).split("\n")
    text = dedent(text)
    out_lines = []
    marker = "$" + tag
    for line_no, line in enumerate(lines):
        line = line.expandtabs(4)
        if marker not in line:
            out_lines.append(line)
            continue
        i = line.find(marker)
        text = indent(text, " "*i)
        out_lines += text.split("\n")
        break
    out_lines += lines[line_no+1:]
    return "\n".join(out_lines)
    
def insert_sql(outer, **tags):
    out = outer
    for tag, value in tags.items():
        out = insert_sql_one(out, tag, value)
    return out




import sys, getopt
from metacat.auth import password_digest_hash
from metacat.ui.cli import CLI, CLICommand

Usage="""
metacat admin -c <config file>  create <username> <password>   - create new admin account
                                password <username> <password> - change admins password
                                add <username>                 - add admin privileges
                                remove <username>              - remove admin privileges
                                list                           - list all admin accounts
                                genkey [-l <length>] [-x]      - generate random key, length in bytes, default 32
                                                                 if -x is used, generate random bytes in hex

Requires direct access to the database. The YAML config file must include:

    database:
        host: ...
        port: ...
        user: ...
        password: ...
        dbname: ...
"""

def connect(config):
    import psycopg2
    dbcfg = config["database"]
    connstr = "host=%(host)s port=%(port)s dbname=%(dbname)s user=%(user)s password=%(password)s" % dbcfg
    conn = psycopg2.connect(connstr)
    schema = dbcfg.get("schema")
    if schema:
        conn.cursor().execute(f"set search_path to {schema}")
    return conn

class ListCommand(CLICommand):

    Usage = """                                     -- list admins"""

    def __call__(self, command, config, opts, args):
        from metacat.db import DBUser
        db = connect(config)
        users = DBUser.list(db)
        for u in users:
            if u.is_admin():
                print(u.Username, u.Name)

class CreateCommand(CLICommand):

    MinArgs = 2
    Usage = """<username> <password>                -- create new admin account"""

    Realm = "metacat"

    def __call__(self, command, config, opts, args):
        from metacat.db import DBUser
        db = connect(config)
        realm = opts.get("-r", config.get("authentication", {}).get("realm"))
        if not realm:
            print("Digest authentication realm must be specified either in the config file (authentication->realm) or with -r",
                file = sys.stderr)
            sys.exit(2)
        username, password = args
        u = DBUser.get(db, username)
        if u is not None:
            print("User already exists. Leaving users status unchanged. Use 'metacat admin add ...'")
            sys.exit(1)
        u = DBUser(db, username, "Admin", "", flags="a")
        u.set_password(self.Realm, password)
        u.save()
        print("Admin user %s created" % (username,))

class PasswordCommand(CLICommand):
    
    MinArgs = 2
    Usage = """[-r <realm>] <username> <password>   -- change account password
        -r <realm>                  -- realm to be used for the digest authentication
                                       can be also specified in the config file:
                                        
                                        authentication:
                                            realm: ...
    """
    Opts = "r:"
    
    def __call__(self, command, config, opts, args):
        from metacat.db import DBUser
        db = connect(config)
        realm = opts.get("-r", config.get("authentication", {}).get("realm"))
        if not realm:
            print("Digest authentication realm must be specified either in the config file (authentication->realm) or with -r",
                file = sys.stderr)
            sys.exit(2)
        username, password = args
        u = DBUser.get(db, username)
        if u is None or not u.is_admin():
            print("User does not exist or is not an Admin. Leaving the password unchanged.")
            sys.exit(1)
        u.set_password(realm, password)
        #print("hashed password:", hashed)
        u.save()
        print("Password updated")

class AddCommand(CLICommand):

    MinArgs = 1
    Usage = """<username>                           -- add admin privileges to an existing account"""

    def __call__(self, command, config, opts, args):
        from metacat.db import DBUser
        username = args[0]
        db = connect(config)
        u = DBUser.get(db, username)
        if u is None or u.is_admin():
            print("User does not exist or is an Admin already.")
            sys.exit(1)
        u.Flags = (u.Flags or "") + "a"
        u.save()
        print("Admin privileges added")

class RemoveCommand(CLICommand):

    MinArgs = 1
    Usage = """<username>                           -- remove admin privileges from an account"""
    
    def __call__(self, command, config, opts, args):
        from metacat.db import DBUser
        username = args[0]
        db = connect(config)
        u = DBUser.get(db, username)
        if u is None or not u.is_admin():
            print("User does not exist or is not an Admin.")
            sys.exit(1)
        u.Flags = (u.Flags or "").replace("a", "")
        u.save()
        print("Admin privileges removed")

class GenerateCommand(CLICommand):
    
    Opts = "xl:"
    Usage = """[-x] [-l <length>]                   -- generate random password
        -x                  -- bytes, print as hex
        -l <length>         -- password length    
    """
    
    
    def __call__(self, command, config, opts, args):
        from metacat.db import DBUser
        import secrets
        length = int(opts.get("-l", 32))
        if "-x" in opts:
            key = secrets.token_hex(length)
        else:
            key = secrets.token_urlsafe(length)
        print (key)
        
class AdminCLI(CLI):

    Opts = "c:"
    Hidden = True
    Usage = "-c <database config YAML file> <command> ..."
    
    def update_context(self, context, command, opts, args):
        import yaml
        if "-c" not in opts:
            print("Database configuration must be specified with -c", file=sys.stderr)
            sys.exit(2)
        cfg = yaml.load(open(opts["-c"], "r"), Loader=yaml.SafeLoader)
        return cfg
 
AdminCLI = AdminCLI(
    "create",       CreateCommand(),
    "add",          AddCommand(),
    "password",     PasswordCommand(),
    "list",         ListCommand(),
    "remove",       RemoveCommand(),
    "generate",     GenerateCommand()
    )






import hashlib, re
from .py3 import to_str, to_bytes
from .password_hash import password_digest_hash

class Authenticator(object):
    
    def __init__(self, config, db_info):
        self.Config = config
        self.Info = db_info        # User authentication info stored in the DB, e.g. hashed password or X.509 DNs

    def authenticate(self, username, presented):
        # info is the DB representation of the secret, e.g. hashed password
        # presented is the user authentication info presented by the client, e.g. unhashed password
        raise NotImplementedError()
        # returns yes/no, info

    def enabled(self):
        return True
        
    def update_auth_info(self, *params, **args):
        raise NotImplementedError()
        # returns updated info

class PasswordAuthenticator(Authenticator):
    
    #
    # passwords are stored as MD5("<username>:<realm>:<password>"), suitbale for RFC2617 digest authentication
    #
    
    def __init__(self, realm, db_info):
        self.DBInfo = (db_info or {}).copy()
        self.Realm = realm
        self.DBHashed = self.DBInfo.get(self.Realm)
    
    def authenticate(self, username, password):
        # will accept both hashed and not hashed password
        return password == self.DBHashed or self.password_hash(username, password) == self.DBHashed, None

    def password_hash(self, username, password):
        return password_digest_hash(self.Realm, username, password).hex().lower()
        
    def enabled(self):
        return self.DBHashed is not None
        
    def update_auth_info(self, username, password, hashed=False):
        hashed_password = password if hashed else self.password_hash(username, password)
        self.DBInfo[self.Realm] = self.DBHashed = hashed_password
        return self.DBInfo

class LDAPAuthenticator(Authenticator):
    
    def authenticate(self, username, password):
        import ldap
        config = self.Config
        if config is None or not "server_url" in config:
            #print("server not configured")
            return False
        dn = self.Info             # LDAP DN
        if not dn and "dn_template" in self.Config:
            dn = self.Config["dn_template"] % (username,)
        if not dn:
            #print("no dn")
            return False        # not allowed

        ld = ldap.initialize(config["server_url"])
        #print("ldap password:", password)
        try:
            ld.simple_bind_s(dn, password)
            result = True
        except ldap.INVALID_CREDENTIALS:
            result = False
        return result, None
        
    def enabled(self):
        return self.Info is not None or "dn_template" in self.Config


class DN(object):

    def __init__(self, inp):
        self.Text = inp
        self.Fields = self.parse(inp)

    def __str__(self):
        return f"DN({self.Text})"

    __repr__ = __str__

    RFCRE=re.compile("(\s*,\s*)?([A-Z]+)=")
    LegacyRE=re.compile("\s*/([A-Z]+)=")
    
    def items(self):
        yield from sorted(self.Fields.items())
    
    __iter__ = items

    @staticmethod
    def parse(text):
        fields = {}
        if text.startswith('/'):
            parts = DN.LegacyRE.split(text)[1:]
            pairs = [(parts[i], parts[i+1]) for i in range(0, len(parts), 2)]
        else:
            parts = DN.RFCRE.split(text)[1:]
            pairs = [(parts[i+1], parts[i+2]) for i in range(0, len(parts), 3)]
        for name, value in pairs:
            name = name.upper()
            if name == "CN":
                # if this is numric CN, e.g. CN=12345, ignore it
                try:    _=int(value)
                except: pass
                else:   continue        # if it's an integer, skip it
            fields.setdefault(name, []).append(value)
        return {name: sorted(lst) for name, lst in fields.items()}

    def __eq__(self, other):
        return self.Fields == other.Fields

    def __ge__(self, other):
        for name, lst in self.Fields.items():
            lst1 = other.Fields.get(name, [])
            if any(v not in lst for v in lst1):
                return False
        else:
            return True

    def __le__(self, other):
        return other >= self

    def legacy(self):
        out = []
        for name, values in self:
            for value in values:
                out.append(f"/{name}={value}")
        return ''.join(out)

    def rfc(self):
        out = []
        for name, values in self:
            for value in values:
                out.append(f"{name}={value}")
        return ','.join(out)



class X509Authenticator(Authenticator):
    
    def authenticate(self, username, request_env):
        known_dns = self.Info or []
        #log = open("/tmp/x509.log", "w")
        #print("known_dns:", known_dns, file=log)
        subject = request_env.get("SSL_CLIENT_S_DN")
        issuer = request_env.get("SSL_CLIENT_I_DN")

        if not subject or not issuer or not known_dns:
            return False

        subject = DN(subject)
        issuer = DN(issuer)
        known_dns = [DN(dn) for dn in known_dns]

        #print("subject:", subject, file=log)
        #print("issuer:", issuer, file=log)
        #print("subject in known_dns:", subject in known_dns, file=log)
        #print("issuer in known_dns:", issuer in known_dns, file=log)
        #print("issuer in subject:", issuer in subject, file=log)
        return (
            any(subject == dn for dn in known_dns) or                     # cert
            any(issuer == dn for dn in known_dns) and issuer <= subject   # proxy
        ), None
        
    def enabled(self):
        return not not self.Info

def authenticator(method, config, user_info):
    if method == "password":   a = PasswordAuthenticator(config, user_info)
    elif method == "x509":     a = X509Authenticator(config, user_info)
    elif method == "ldap":     a = LDAPAuthenticator(config, user_info)
    else:
        raise ValueError(f"Unknown autenticator type {method}")
    return a
    


import os, os.path, stat
from .signed_token_jwt import SignedToken, SignedTokenExpiredError, SignedTokenImmatureError
from .py3 import to_bytes, to_str

class TokenLib(object):

        def __init__(self, path = None, create = True):
            if not path:
                locations = [
                    f"{location}/.token_library" for location in 
                    [ os.environ.get("HOME"), os.getcwd(), "/tmp" ]
                    if location
                ]
            else:
                locations = [path]
            self.Tokens, self.Location = self.load_library(locations)
            if not self.Location and create:
                # not found
                self.Location = self.create_library(locations)
                
        def create_library(self, paths):
            for path in paths:
                try:    
                    open(path, "w").close()         # create empty library
                    os.chmod(path, stat.S_IRUSR | stat.S_IWUSR )
                    return path
                except: pass
            return None

        def load_library(self, paths):
            #print("load_library: paths:", paths)
            for path in paths:
                if os.path.isfile(path):
                    try:    
                        #print("loading tokens from:", path)
                        tokens = self.load_from_file(path)
                        #print("tokens loaded from:", path)
                        return tokens, path
                    except:
                        pass
            return {}, None     # not found
            
        def load_from_file(self, path):
            lines = open(path, "r").readlines()
            out = {}
            for line in lines:
                line = line.strip()
                url, encoded = line.split(None, 1)
                try:
                    token = SignedToken.decode(encoded)
                    #print("TokenLib.load: token:", token)
                    token.verify()      # this will verify expiration/maturity times only
                    #print("  token verified. Exp:", token.Expiration)
                except SignedTokenExpiredError:
                    #print("TokenLib.load: token expired")
                    token = None
                except SignedTokenImmatureError:
                    #print("TokenLib.load: token immature")
                    pass
                if token is not None:
                    out[url] = token
            #print("TokenLib.load: out:", out)
            return out

        def save_tokens(self):
            if self.Location:
                f = open(self.Location, "w")
                for url, token  in self.Tokens.items():
                    f.write("%s %s\n" % (url, to_str(token.encode())))
                f.close()

        def __setitem__(self, url, token):
                if isinstance(token, (str, bytes)):
                    token = SignedToken.decode(token)
                self.Tokens[url] = token
                self.save_tokens()

        def __getitem__(self, url):
                return self.Tokens[url]

        def get(self, url):
                return self.Tokens.get(url)

        def items(self):
                return self.Tokens.items()
                
        def exists(self):
            return not not self.Location


from .py3 import PY2, PY3, to_str, to_bytes
from .timelib import epoch
from .trace import Tracer
from .object_spec import ObjectSpec, undid
from .utils import chunked, limited, unique, first_not_empty, strided, skipped, insert_sql
from .validation import validate_metadata



from metacat.common.trees  import Ascender, Node
from metacat.db import DBFileSet
from .meta_evaluator import MetaEvaluator

class FileQueryExecutor(Ascender):
    
    def __init__(self, db, filters, debug=False):
        self.DB = db
        self.Filters = filters
        self.Debug = False
        
    def debug(self, *params, **args):
        if self.Debug:
            parts = ["FileQueryExecutor:"]+list(params)
            print(*parts, **args)
            
    def __call__(self, tree):
        result = self.walk(tree)
        return result
        
    def empty(self, node, *args):
        return DBFileSet(self.DB)    # empty file set
        
    def sql(self, node, sql=None):
        #print("sql:", sql)
        return DBFileSet.from_sql(self.DB, sql)
        
    def meta_filter(self, node, query=None, meta_exp=None, with_meta=False, with_provenance=False):
        #print("meta_filter: args:", args)
        #print("meta_filter: node:", node, "  query:", query)
        evaluator = MetaEvaluator()
        out = (f for f in query
                if evaluator(f.metadata(), meta_exp)
        )
        return DBFileSet(self.DB, out)

    def union(self, node, *args):
        return DBFileSet.union(self.DB, args)

    def join(self, node, *args, **kv):
        return DBFileSet.join(self.DB, args)
        
    def ordered(self, node):
        assert isinstance(node, DBFileSet)
        return node.ordered()

    def minus(self, node, *args, **kv):
        #print("Evaluator.union: args:", args)
        assert len(args) == 2
        assert all(n.T in ("sql", "file_set", "empty") for n in args)
        left, right = args
        return left - right

    def parents_of(self, node, arg, with_meta=False, with_provenance=False):
        return arg.parents(as_files=True, with_metadata=with_meta, with_provenance=with_provenance)

    def children_of(self, node, *args, with_meta=False, with_provenance=False):
        return arg.children(as_files=True, with_metadata=with_meta, with_provenance=with_provenance)

    def skip_limit(self, node, arg, skip=0, limit=None, **kv):
        return arg.skip(skip).limit(limit)
            
    def filter(self, node, *queries, name=None, params=[], kw={}, 
                skip=0, limit=None,
                with_meta=False, ordered=False):
        #print("Evaluator.filter: inputs:", inputs)
        assert name is not None
        filter_object = self.Filters[name]
        #print("filter: queries:", queries)
        return DBFileSet(self.DB, filter_object.run(queries, params, kw, 
                limit=limit, skip=skip, with_meta=with_meta, ordered=ordered))



def validate_metadata(definitions, restricted, metadata={}, name=None, value=None):
    """
    Validates metadata against a set of definitions. Can be used to validate category and dataset restrictions
    
    Parameters
    ----------
    metadata : dict
        metadata dictionary to validate
    kw :
        additional metadata names and values
    definitions : dict
        metadata definitions, indexed by the parameter name:
            {
                "parameter_name" {
                    ... definition ...
                }
            }
        
        definition may include:
            type : string, one of: int, float, text, boolean, dict, list, int[], float[], text[], boolean[], any
            min : minimal acceptable value (for numeric and strings)
            max : max acceptable value (for numeric and strings)
            values : list, enumeration of values
            pattern : regexp pattern for text strings
            required : boolean, whether the parameter is required
    restricted : boolean
        whether the validation must be done in "restricted" mode - if a parameter is not mentioned in the definitions, flag that as an error

    Returns
    -------
    list
        list of tuples (name, error) - error mesages for parameters
        empty list indicates that validation succeeded
    """

    definitions = definitions or {}
    
    metadata = (metadata or {}).copy()
    if name is not None:
        metadata[name] = value

    errors = []
    for name, value in metadata.items():

        definition = definitions.get(name)
        if definition is None:    
            if restricted:
                errors.append((name, "parameter not allowed in restricted category"))
            continue

        typ = definition.get("type")
        type_mismatch = False
        repv = repr(value)
        if typ is not None:
            if typ == "any":
                continue
        
        
            if typ == "int" and not isinstance(value, int): 
                errors.append((name, f"scalar int value required instead of {repv}"))
                type_mismatch = True
            elif typ == "float" and not isinstance(value, float): 
                errors.append((name, f"scalar float value required instead of {repv}"))
                type_mismatch = True
            elif typ == "text" and not isinstance(value, str): 
                errors.append((name, f"scalar text value required instead of {repv}"))
                type_mismatch = True
            elif typ == "boolean" and not isinstance(value, bool): 
                errors.append((name, f"scalar boolean value required instead of {repv}"))
                type_mismatch = True
            elif typ == "dict" and not isinstance(value, dict): 
                errors.append((name, f"dict value required instead of {repv}"))
                type_mismatch = True
            elif typ == "list" and not isinstance(value, list): 
                errors.append((name, f"list value required instead of {repv}"))
                type_mismatch = True

            elif typ == "int[]":
                if not isinstance(value, list): 
                    errors.append((name, f"list of ints required instead of {repv}"))
                    type_mismatch = True
                if not all(isinstance(x, int) for x in value): 
                    errors.append((name, f"list of ints required instead of {repv}"))
                    type_mismatch = True

            elif typ == "float[]":
                if not isinstance(value, list): 
                    errors.append((name, f"list of floats required"))
                    type_mismatch = True
                if not all(isinstance(x, float) for x in value): 
                    errors.append((name, f"list of floats required instead of {repv}"))
                    type_mismatch = True
            
            elif typ == "text[]":
                if not isinstance(value, list): 
                    errors.append((name, f"list of strings required"))
                    type_mismatch = True
                if not all(isinstance(x, str) for x in value): 
                    errors.append((name, f"list of strings required instead of {repv}"))
                    type_mismatch = True
            
            elif typ == "boolean[]":
                if not isinstance(value, list): 
                    errors.append((name, f"list of booleans required"))
                    type_mismatch = True
                if not all(isinstance(x, bool) for x in value): 
                    errors.append((name, f"list of booleans required instead of {repv}"))
                    type_mismatch = True
        
        if not type_mismatch:
            if not typ in ("boolean", "boolean[]", "list", "dict", "any"):
                if "values" in definition:
                    values = definition["values"]
                    if isinstance(value, list):
                        if not all(x in values for x in value): errors.append((name, f"value in {value} is not allowed"))
                    else:
                        if not value in values: errors.append((name, f"value {repv} is not allowed"))
                else:
                    if "pattern" in definition and typ in ("text", "text[]"):
                        pattern = definition["pattern"]
                        r = re.compile(pattern)
                        if isinstance(value, list):
                            if not all(r.match(v) is not None for v in value):  
                                errors.append((name, f"value in {value} does not match the pattern '{pattern}'"))
                        else:
                            if r.match(value) is None:
                                errors.append((name, f"value {value} does not match the pattern '{pattern}'"))
                    if "min" in definition:
                        vmin = definition["min"]
                        if isinstance(value, list):
                            if not all(x >= vmin for x in value):   
                                errors.append((name, f"value in {value} out of range (min:{vmin})"))
                        else:
                            if value < vmin:    
                                errors.append((name, f"value {value} out of range (min:{vmin})"))
                    if "max" in definition:
                        vmax = definition["max"]
                        if isinstance(value, list):
                            if not all(x <= vmax for x in value):   errors.append((name, f"value in {value} out of range (max:{vmax})"))
                        else:
                            if value > vmax:    errors.append((name, f"value {value} out of range (max:{vmax})"))

    for dname, definition in definitions.items():
        if definition.get("required") and dname not in metadata:
            errors.append((dname, "required parameter is missing"))
    
    return errors


from .http_client import HTTPClient
from .exceptions import MCError, NotFoundError, ServerSideError, InvalidArgument, PermissionError, BadRequestError, WebAPIError
from .meta_dnf import MetaExpressionDNF



import requests, time, json, random
from urllib.parse import quote_plus, unquote_plus
from .exceptions import MCError, NotFoundError, ServerSideError, InvalidArgument, PermissionError, BadRequestError, WebAPIError

def to_bytes(x):
    if not isinstance(x, bytes):
        x = x.encode("utf-8")
    return x
    
def to_str(x):
    if isinstance(x, bytes):
        x = x.decode("utf-8")
    return x


class HTTPClient(object):

    InitialRetry = 1.0
    RetryExponent = 1.5
    DefaultTimeout = 300.0

    def __init__(self, server_url, token, timeout):
        self.ServerURL = server_url
        self.Token = token
        self.Timeout = timeout or self.DefaultTimeout
        self.LastResponse = self.LastURL = self.LastStatusCode = None

    def retry_request(self, method, url, timeout=None, **args):
        """
        Implements the functionality to retry on 503 response with random exponentially growing delay
        Use timemout = 0 to try the request exactly once
        Returns the response with status=503 on timeout
        """
        if timeout is None:
            timeout = self.DefaultTimeout
        tend = time.time() + timeout
        retry_interval = self.InitialRetry
        response = None
        done = False
        while time.time() < tend:
            if method == "get":
                response = requests.get(url, timeout=self.Timeout, **args)
            else:
                response = requests.post(url, timeout=self.Timeout, **args)
            if response.status_code != 503:
                break
            sleep_time = min(random.random() * retry_interval, tend-time.time())
            retry_interval *= self.RetryExponent
            if sleep_time > 0:
                time.sleep(sleep_time)
        return response
        
    def raise_on_error(self, response):
        if response.status_code == 403:
            raise PermissionError(url, response)
        elif response.status_code == 404:
            raise NotFoundError(url, response)
        elif response.status_code//100 == 4:
            raise BadRequestError(url, response)
        elif response.status_code//100 != 2:
            raise WebAPIError(url, response)

    def send_request(self, method, uri_suffix, headers=None, timeout=None, **args):
        self.LastURL = url = "%s/%s" % (self.ServerURL, uri_suffix)
        req_headers = {
            "Accept": "text/plain, application/json, text/json, application/json-seq"
        }
        try:
            req_headers.update(self.auth_headers())           # in case we have TokenAuthClientMixin or similar
        except AttributeError:
            pass
        if headers:
            req_headers.update(headers)
        self.LastResponse = response = self.retry_request(method, url, headers=req_headers, **args)
        #print(response, response.text)
        self.LastStatusCode = response.status_code
        self.raise_on_error(response)
        return response

    def get_text(self, uri_suffix):
        return self.send_request("get", uri_suffix).text

    def post_text(self, uri_suffix, data):
        return self.send_request("post", uri_suffix, data=data).text

    def unpack_json(self, json_text):
        results = json.loads(json_text)
        #if isinstance(results, dict):
        #    if "results" in results:
        #        results = results["results"]
        #    elif "error" in results:
        #        raise ServerReportedError(self.LastURL, self.LastStatusCode, results["error"]["type"], results["error"].get("value", ""))
        return results

    RS = b'\x1E'
    def unpack_json_seq(self, response):
        for line in response.iter_lines():
            if line:    line = line.strip()
            while line and line.startswith(self.RS):
                line = line[1:]
            if line:
                #print(f"stream line:[{line}]")
                obj = json.loads(line)
                yield obj

    def unpack_json_data(self, response):
        response_content_type = response.headers.get("content-type")
        if "application/json-seq" in response_content_type:
            return self.unpack_json_seq(response)
        else:
            return response.json()

    def get_json(self, uri_suffix):
        headers = {"Accept": "application/json-seq, application/json, text/json"}
        return self.unpack_json_data(self.send_request("get", uri_suffix, headers=headers, stream=True))

    def post_json(self, uri_suffix, data):
        if not isinstance(data, (str, bytes)):
            data = json.dumps(data)
        headers = {
                "Accept": "application/json-seq, application/json, text/json",
                "Content-Type": "text/json"
        }
        response = self.send_request("post", uri_suffix, data=data, headers=headers, stream=True)
        return self.unpack_json_data(response)

    def get_json_stream(self, uri_suffix):
        url = "%s/%s" % (self.ServerURL, uri_suffix)
        headers = {"Accept": "application/json-seq"}
        if self.Token is not None:
            headers["X-Authentication-Token"] = self.Token.encode()

        with self.retry_request("get", url, headers=headers, stream=True) as response:
            if response.status_code == INVALID_METADATA_ERROR_CODE:
                raise InvalidMetadataError(url, response.status_code, response.text)
            if response.status_code == 404:
                raise NotFoundError(url, response.status_code, response.text)
            elif response.status_code != 200:
                raise WebAPIError(url, response.status_code, response.text)
            
            if response.headers.get("Content-Type") != "application/json-seq":
                raise WebAPIError(url, 200, "Expected content type application/json-seq. Got %s instead." % (response.headers.get("Content-Type"),))

            for line in response.iter_lines():
                if line:    line = line.strip()
                while line.startswith(b'\x1E'):
                    line = line[1:]
                if line:
                    #print(f"stream line:[{line}]")
                    obj = json.loads(line)
                    yield obj

    def interpret_json_stream(self, response):
        for line in response.iter_lines():
            if line:    line = line.strip()
            while line.startswith(b'\x1E'):
                line = line[1:]
            if line:
                #print(f"stream line:[{line}]")
                obj = json.loads(line)
                yield obj
                    
    def interpret_response(self, response, none_if_not_found=True):
        if response.status_code != 200:
            if none_if_not_found and response.status_code == 404:
                return None
            elif response.status_code == 404:
                raise NotFoundError(self.LastURL, response)
            else:
                raise WebAPIError(self.LastURL, response)
        content_type = response.headers.get("Content-Type", "")
        if content_type.startswith("text/json"):
            return json.loads(response.text)
        elif content_type == "application/json-seq":
            return self.interpret_json_stream(response)
        else:
            return response.text

    def get(self, uri_suffix, none_if_not_found=False):
        if not uri_suffix.startswith("/"):  uri_suffix = "/"+uri_suffix
        url = "%s%s" % (self.ServerURL, uri_suffix)
        #print("url:", url)
        try:
            headers = self.auth_headers()           # in case we have TokenAuthClientMixin or similar
        except AttributeError:
            headers = {}
        response = self.retry_request("get", url, headers=headers)
        return self.interpret_response(response, none_if_not_found)

    def post(self, uri_suffix, data):
        #print("post_json: data:", type(data), data)
        
        if not uri_suffix.startswith("/"):  uri_suffix = "/"+uri_suffix
        
        if data is None or isinstance(data, (dict, list)):
            data = json.dumps(data)
        else:
            data = to_bytes(data)
        #print("post_json: data:", type(data), data)
            
        url = "%s%s" % (self.ServerURL, uri_suffix)
        
        try:
            headers = self.auth_headers()           # in case we have TokenAuthClientMixin or similar
        except AttributeError:
            headers = {}

        response = self.retry_request("post", url, data=data, headers=headers)
        return self.interpret_response(response)



import sys, getopt, os, json, pprint
from urllib.request import urlopen, Request
from urllib.parse import quote_plus, unquote_plus
from metacat.util import to_bytes, to_str
from metacat.webapi import MetaCatClient, MCServerError, MCWebAPIError, MCError
from metacat.ui.cli import CLICommand, InvalidArguments, InvalidOptions

Usage = """
Usage: --
    metacat query <options> "<MQL query>"
    metacat query <options> -f <MQL query file>

    Options:
        -j|--json                           - print raw JSON output
        -p|--pretty                         - pretty-print metadata
        -l|--line                           - print all metadata on single line (good for grepping, ignored with -j and -p)
        -i|--ids                            - print file ids instead of names
        -s|--summary                        - print only summary information
        -m|--metadata=[<field>,...]         - print metadata fields
                                              overrides --summary
        -m|--metadata=all                   - print all metadata fields
                                              overrides --summary
        -P|--with-provenance                - include provenance information
        -N|--namespace=<default namespace>  - default namespace for the query
        -S|--save-as=<namespace>:<name>     - save files as a new datset
        -A|--add-to=<namespace>:<name>      - add files to an existing dataset
"""

class QueryCommand(CLICommand):

    GNUStyle = False    
    Opts = (
        "jism:N:pq:S:A:lPxr", 
        ["line", "json", "ids", "summary", "metadata=", "namespace=", "pretty",
            "with-provenance", "save-as=", "add-to=", "explain", "include-retired-files"
        ]
    )
    Usage = """[<options>] (-q <MQL query file>|"<MQL query>")

        options:
            -j|--json                           - print raw JSON output
            -p|--pretty                         - pretty-print metadata
            -l|--line                           - print all metadata on single line (good for grepping, ignored with -j and -p)
            -i|--ids                            - print file ids instead of names
            -s|--summary                        - print only summary information
            -m|--metadata [<field>,...]         - print metadata fields
                                                  overrides --summary
            -m|--metadata all                   - print all metadata fields
                                                  overrides --summary
            -P|--with-provenance                - include provenance information
            -N|--namespace=<default namespace>  - default namespace for the query
            -S|--save-as=<namespace>:<name>     - save files as a new datset
            -A|--add-to=<namespace>:<name>      - add files to an existing dataset
            -r|--include-retired-files          - include retired files into the query results

            -x|--explain                        - dp not run the query, show resulting SQL only
    """
    
    def __call__(self, command, client, opts, args):
        namespace = opts.get("-N") or opts.get("--namespace")
        #with_meta = not "--summary" in opts and not "-s" in opts
        with_meta = "-m" in opts or "--metadata" in opts
        with_provenance = "-P" in opts or "--with-provenance" in opts
        keys = opts.get("-m") or opts.get("--metadata") or []
        if keys and keys != "all":    keys = keys.split(",")
        save_as = opts.get("-S") or opts.get("--saves-as")
        add_to = opts.get("-A") or opts.get("--add-to")
        include_retired = "-r" in opts or "--include-retired-files" in opts

        #print("url:", url)
        if args:
            query_text = " ".join(args)
        else:
            query_file = opts.get("-q")
            if not query_file:
                raise InvalidArguments("Query must be specified")
            query_text = to_str(open(query_file, "r").read())
            
        if "-x" in opts or "--explain" in opts:
            from metacat.mql import MQLQuery
            print("---- Query text ----\n%s\n" % (query_text,))
            q = MQLQuery.parse(query_text, loader=client, include_retired_files=include_retired)

            print("---- Parsed ----")
            print(q.Parsed.pretty())
            print("")

            print("---- Converted ----")
            print(q.Tree.pretty("    "))
            print("")
        
            #q.optimize(False)
            #print("---- Optimized ----")
            #print(q.Optimized.pretty("    "))
            #print("")
    
            compiled = q.compile(with_meta=with_meta, with_provenance=with_provenance)
            print("---- Compiled ----")
            print(compiled.pretty("    "))
        else:
            results = client.query(query_text, 
                        namespace=namespace, with_metadata = with_meta, 
                        save_as=save_as, add_to=add_to,
                        with_provenance=with_provenance,
                        include_retired_files=include_retired
            )
            if "--json" in opts or "-j" in opts:
                results = list(results)
                print(json.dumps(results, sort_keys=True, indent=4, separators=(',', ': ')))
                sys.exit(0)

            if "--pretty" in opts or "-p" in opts:
                results = list(results)
                meta = sorted(results, key=lambda x: x["name"])
                pprint.pprint(meta)
                sys.exit(0)

            in_line = "-l" in opts or "--line" in opts

            #print("response results:", results)
    
            if "-s" in opts or "--summary" in opts and not with_meta:
                nfiles = total_size = 0
                for f in results:
                    nfiles += 1
                    total_size += f.get("size", 0)
                print("Files:       ", nfiles)
                if total_size >= 1024*1024*1024*1024:
                    unit = "TB"
                    n = total_size / (1024*1024*1024*1024)
                elif total_size >= 1024*1024*1024:
                    unit = "GB"
                    n = total_size / (1024*1024*1024)
                elif total_size >= 1024*1024:
                    unit = "MB"
                    n = total_size / (1024*1024)
                elif total_size >= 1024:
                    unit = "KB"
                    n = total_size / 1024
                else:
                    unit = "B"
                    n = total_size
                print("Total size:  ", "%d (%.3f %s)" % (total_size, n, unit))
                    
            else:
                for f in results:
                    meta_lst = []
                    meta_out = ""
                    if with_meta:
                        meta = f["metadata"]
                        klist = sorted(list(meta.keys())) if keys == "all" else keys
                        if in_line:
                            for k in klist:
                                if k in meta:
                                    meta_lst.append("%s=%s" % (k, repr(meta[k])))
                        else:
                            for k in klist:
                                if k in meta:
                                    meta_lst.append("%s\t=\t%s" % (k, repr(meta[k])))
                        if meta_lst:
                            if in_line:
                                meta_out = "\t"+"\t".join(meta_lst)
                            else:
                                meta_out += "\n    "+"\n    ".join(meta_lst)
                    if "--ids" in opts or "-i" in opts:
                        print("%s%s" % (f["fid"],meta_out))
                    else:
                        print("%s:%s%s" % (f["namespace"], f["name"], meta_out))

QueryInterpreter = QueryCommand()


from lark import Lark
from lark import Transformer, Tree, Token
import pprint
from textwrap import dedent

class SyntaxTreeConversionError(Exception):
    pass

class Token(object):
    
    JSON_CLASS = "token"

    def __init__(self, typ, value):
        # print("Token created:", typ, value)
        self.T = typ
        self.V = value

    def __str__(self):
        return "Token(%s, %s)" % (self.T, self.V)

    def pretty(self):
        #print("pretty---")
        return self._pretty()
        
    def _pretty(self, indent="", headline_indent=None):
        return 'token %s "%s"' % (self.T, self.V), []

    def jsonable(self):
        return dict(T=self.T, V=self.V)
        
    def to_json(self):
        d = self.jsonable()
        d["///class///"] = self.JSON_CLASS
        return json.dumps(d)

    @staticmethod
    def from_jsonable(data):
        if isinstance(data, dict) and data.get("///class///") == Token.JSON_CLASS:
            typ = data["T"]
            val = data["V"]
            return Token(typ, val)
        else:
            return data

    @staticmethod
    def from_json(text):
        return Token.from_jsonable(json.loads(text))

class Node(object):

    JSON_CLASS = "node"

    def __init__(self, typ, children=[], _data=None, _meta=None, **kw):
        self.T = typ
        self.M = _meta
        self.C = children[:]
        self.D = {}
        self.D.update(_data or {})
        self.D.update(kw)

    def __str__(self):
        return "Node(%s (%d children) (%d data) meta:%s)" % (self.T, len(self.C), len(self.D), self.M)
        
    __repr__ = __str__
    
    def __getitem__(self, name):
        return self.D[name]

    def clone(self, children=None, **kw):
        if children is None:    children = self.C[:]
        d = {}
        d.update(self.D)
        d.update(kw)
        return Node(self.T, children, _data=d, _meta=self.M)
        
    def get(self, name, default=None):
        return self.D.get(name, default)

    def __setitem__(self, name, value):
        self.D[name] = value
        
    def _pretty(self, indent=""):
        out = []
        items = list(self.D.items())
        nitems = len(items)
        nc = len(self.C)
        for i, (k, v) in enumerate(items):
            key = f"{k} = "
            if isinstance(v, (Node, Token)):
                key_len = len(key)
                prefix = "|  " if i < len(items) - 1 else "   "
                head, lines = v._pretty(indent + prefix + " "*key_len + " ")
                out.append(indent + "+   " + key + head)
                out += lines
            else:
                str_v = dedent(str(v))
                lines = str_v.split("\n")
                if len(lines) > 1:
                    out.append(indent + "+   " + key)
                    prefix = "|   " if i < nitems-1 or nc else "    "
                    for l in lines:
                        out.append(indent + prefix + " " + l)
                else:
                    out.append(indent + "+   " + key + str(v))

        for i, c in enumerate(self.C):
            child_indent = indent + (' ' if i == nc-1 else '|') + "  "
            if isinstance(c, (Node, Token)):
                head, lines = c._pretty(child_indent)
                out.append(indent + "+- " + head)
                out.extend(lines)
            else:
                out.append(indent + "+- " + str(c))

        head = self.T
        if self.M is not None:
            head += f" m:{self.M}"

        return head, out

    def pretty(self, indent=""):
        #print("pretty---")
        head, lines = self._pretty(indent)
        return indent + head + "\n" + "\n".join(lines)
        
    def jsonable(self):
        d = dict(T=self.T, M=self.M, C=[c.jsonable() if isinstance(c, Node) else c
                        for c in self.C]
        )
        return d
        
    def to_json(self):
        d = self.jsonable()
        d["///class///"] = self.JSON_CLASS
        return json.dumps(d)

    @staticmethod
    def from_jsonable(data):
        if isinstance(data, dict) and data.get("///class///") == Node.JSON_CLASS:
            typ = data["T"]
            if typ == "DataSource":
                return DataSource.from_jsonable(data)
            elif typ == "MetaExp":
                return MetaExp.from_jsonable(data)
            else:
                return Node(data["T"],
                    children = [Node.from_jsonable(c) for c in data.get("C", [])],
                    meta = data.get("M")
            )
        else:
            return data

    @staticmethod
    def from_json(text):
        return Node.from_jsonable(json.loads(text))
        
    def find_all(self, node_type=None, predicate=None, top_down=True):
        def match(c):
            return (node_type is not None and c.T == node_type) or (predicate is not None and predicate(c))
        if top_down and match(self):
            yield self
        for c in self.C:
            if match(c):
                yield(c)
        if not top_down and match(self):
            yield self

def pass_node(method):
    def decorated(self, *params, **args):
        return method(self, *params, **args)
    decorated.__pass_node__ = True
    return decorated

class Visitor(object):	# deprecated

    #
    # Visits each node top-down. Calls corresponding method of each node passing the "context". 
    # If the method returns True, recursively visits the node children
    # Can be used to re-calculate metadata
    #

    def walk(self, node, context=None):
        if not isinstance(node, Node):
            return
        node_type, children = node.T, node.C

        try:
            if hasattr(self, node_type):
                method = getattr(self, node_type)
                visit_children = method(node, context)
            else:
                visit_children = self._default(node, context)
        except Exception as e:
            raise SyntaxTreeConversionError(f"Error while processing node {node_type} with children {children}") from e

        if visit_children:
            for c in children:
                self.walk(c, context)
        return node

    def visit_children(self, node, context):
        for c in node.C:
            self.walk(c, context)
        
    def _default(self, node, context):
        return True
        
class Traveler(object):

    def indent(self, text, rel_level=0, level=None, pad='  ', indent=None):
        if level is None:
            level = self.WalkLevel+rel_level          # concrete subclass must set it
        if indent is None:
            indent = level * pad
        min_indent = None
        lines = text.split("\n")
        for line in lines:
            if line:
                line = line.expandtabs()
                l = len(line) - len(line.lstrip())
                if min_indent is None or l < min_indent:
                    min_indent = l
        #print("min_indent:", min_indent)
        if min_indent:
            out_lines = []
            for line in lines:
                out_lines.append(indent + line[min_indent:])
                #print(f"line -> [{line}]")
            text = '\n'.join(out_lines)
        return text

    def __call__(self, *params, **args):
        return self.walk(*params, **args)

class Descender(Traveler):

    #
    # Descends nodes top to bottom, possibly replacing them
    # If a user method is defined for the node type, it has to explicitly call visit_children(self, context)
    # If the user method returns None, it is equivalent to returning the node itself
    # Default method does visit children
    #

    def walk(self, tree, context=None, level=0):
        self.WalkLevel = level
        try:
            return self._walk(tree, context)
        finally:
            self.WalkLevel = level

    def _walk(self, node, context=None):
        self.WalkLevel += 1
        try:
            if not isinstance(node, Node):
                return node

            node_type = node.T
        
            try:
                if hasattr(self, node_type):
                    method = getattr(self, node_type)
                    new_node = method(node, context)
                else:
                    new_node = self._default(node, context)
            except Exception as e:
                raise SyntaxTreeConversionError(f"Error while processing node {node_type}") from e
            
            return new_node

        finally:
            self.WalkLevel -= 1

    def visit_children(self, node, context):
        node.C = [self._walk(c, context) for c in node.C]
        node.D = {
            key:self._walk(n, context)
            for key, n in node.D.items()
        }
        return node

    def _default(self, node, context):
        #print("Descender._default: node:", node.pretty())
        return self.visit_children(node, context)
        
class Ascender(Traveler):
    
    def walk(self, tree, debug=False, level=0):
        self.WalkLevel = level
        try:
            return self._walk(tree, debug)
        finally:
            self.WalkLevel = level
            
    def _walk(self, node, debug=False):
        self.WalkLevel += 1
        try:
            if not isinstance(node, Node):
                return node
            node_type, children = node.T, node.C
            assert isinstance(node_type, str)

            method = self._default
            pass_node = False

            if hasattr(self, node_type):
                method = getattr(self, node_type)
                if hasattr(method, "__pass_node__") and getattr(method, "__pass_node__"):
                    pass_node = True
        
            if pass_node:
                out = method(node)
            else:
                named_children = {
                    name:(self.walk(c, debug) if isinstance(c, Node) else c) 
                    for name, c in node.D.items()
                }
                children = [self._walk(c, debug) for c in children]
                node = node.clone(children)
                node.D = named_children
                out = method(node, *children, **named_children)
                if debug:
                    me = self.__class__.__name__
                    #print(f"{me}: method {node_type} returned:", out.pretty("      ") if isinstance(out, Node) else out)
        finally:
            self.WalkLevel -= 1
        return out
        
    def _default(self, node, *children, **named):
        return Node(node.T, children, _meta=node.M, _data=named)
        
class Converter(Transformer):
    
    #
    # Base class for converters from Lark tree to Node tree
    #
    
    def __call__(self, tree):
        return self.transform(tree)

class LarkToNodes(Converter):
    
    #
    # Converts from Tree structure returned by Lark to Nodes
    #
    
    def __default__(self, data, children, meta):
        #print("LarkToNodes.__default(", data, children, meta, ")")
        return Node(data, children, meta=meta)
        
    def __default_token__(self, token):
        #print("LarkToNodes.__default_token__: token:", token)
        return Token(token.type, token.value)

        
if __name__ == "__main__":
    
    x = Node("leaf", [], value="hello there", meta=(1,2,3))
    y = Node("leaf", [], value=5)
    
    pi = Node("float", 
        [
            Node("sub_float", [], value=1234),
            Node("sub_float", [], meta=1234)
        ], value=3.14)
    e = Node("float", 
        [
            Node("sub_float", [], value=1234),
            Node("sub_float", [], meta=1234)
        ], value=(2.718, 2818)
    )
    
    tree = Node("top", [x,y], value="some value", data=dict(
                pi = pi, e_extended_name = e, x = ("just","a","tuple")
    ))

    print(tree.pretty())
    



#
# not used
#

from metacat.webapi import MCError
import sys

def catch_mc_errors(method):
    def decorated(*params, **args):
        try:
            return method(*params, **args)
        except MCError as e:
            print(e, file=sys.stderr)
            sys.exit(1)
    return decorated


import sys, getopt, os, json, fnmatch, pprint
from datetime import datetime, timezone
from textwrap import indent
#from urllib.request import urlopen, Request
from metacat.util import to_bytes, to_str, epoch
from metacat.webapi import MetaCatClient, MCError

from metacat.ui.cli import CLI, CLICommand, InvalidArguments

Usage = """
Usage: 
    metacat dataset <command> [<options>] ...
    
    Commands and options:
    
        list [<options>] [[<namespace pattern>:]<name pattern>]
            -l|--long           - detailed output
            -c|--file-counts    - include file counts if detailed output
            
        create [<options>] <namespace>:<name> [<description>]
            -M|--monotonic
            -F|--frozen
            -m|--metadata '<JSON expression>'
            -m|--metadata @<JSON file>
            
        add <parent dataset namespace>:<parent name> <child dataset namespace>:<child name> [<child dataset namespace>:<child name> ...]

        remove <parent namespace>:<parent name> <child namespace>:<child name> 
        
        show [<options>] <namespace>:<name>
            -j|--json       - print as JSON
            -p|--pprint     - Python pprint
        
        update <options> <namespace>:<name> [<description>]
            -M|--monotonic (yes|no) - set/reset monotonic flag
            -F|--frozen (yes|no)    - set/reset monotonic flag
            -r|--replace            - replace metadata, otherwise update
            -m|--metadata @<JSON file with metadata> 
            -m|--metadata '<JSON expression>' 
"""

class ListDatasetFilesCommand(CLICommand):
    
    Opts = "mjr with-metadata include-retired-files"
    Usage = """[<options>] <dataset namespace>:<dataset name>           -- list dataset files
        -m|--with-metadata              - include file metadata
        -r|--include-retired-files      - include retired files
        -j                              - as JSON
    """
    MinArgs = 1
    
    def __call__(self, command, client, opts, args):
        dataset_did = args[0]
        with_meta = "-m" in opts or "--with-metadata" in opts
        files = client.get_dataset_files(dataset_did,
                    with_metadata = with_meta,
                    include_retired_files = "-r" in opts or "--include-retired-files" in opts)
        if "-j" in opts:
            first = True
            print("[")
            for f in files:
                if not first:   print(",")
                print(json.dumps(f, indent=2, sort_keys=True), end="")
                first = False
            print("\n]")
        else:
            for f in files:
                print(f"%s:%s" % (f["namespace"], f["name"]))

class ListDatasetsCommand(CLICommand):
    
    Opts = ("lc", ["--long", "--file-counts"])
    Usage = """ [<options>] [<namespace pattern>:<name pattern>]      -- list datasets
            -l|--long           - detailed output
            -c|--file-counts    - include file counts if detailed output
            """
    
    def __call__(self, command, client, opts, args):
        if args:
            patterns = args[0]
        else:
            patterns = "*:*"
        
        if not ':' in patterns:
            raise InvalidArguments()

        ns_pattern, name_pattern = patterns.split(':', 1)
            
        verbose = "-l" in opts or "--long" in opts
        include_counts = verbose and ("-c" in opts or "--file-counts" in opts)
        output = list(client.list_datasets(ns_pattern, name_pattern, with_counts=include_counts))
        output = sorted(output, key=lambda ds:(ds["namespace"], ds["name"]))
    
        if include_counts:
            verbose_format = "%-16s %-23s %10s %5s/%-5s %s"
            header_format = "%-16s %-23s %-10s %-11s %s"
            divider = " ".join(("-"*16, "-"*23, "-"*10, "-"*11, "-"*60))
            columns = ("creator", "created", "files", "subsets", "namespace:name")
        else:
            verbose_format = "%-16s %-23s %s"
            header_format = "%-16s %-23s %s"
            divider = " ".join(("-"*16, "-"*23, "-"*60))
            columns = ("creator", "created", "namespace:name")
            
        if verbose:
            print(header_format % columns)
            print(divider)
    
        for item in output:
            match = False
            namespace, name = item["namespace"], item["name"]
            if fnmatch.fnmatch(name, name_pattern) and fnmatch.fnmatch(namespace, ns_pattern):
                if verbose:
                    ct = item.get("created_timestamp")
                    if not ct:
                        ct = ""
                    else:
                        ct = datetime.fromtimestamp(ct, timezone.utc).strftime("%Y-%m-%d %H:%M:%S %Z")
                    file_count = item.get("file_count")
                    if file_count is None:
                        file_count = "?"
                    else:
                        file_count = str(file_count)
                    child_count = item.get("child_count", "?")
                    subset_count = item.get("subset_count", "?")
                    if include_counts:
                        print(verbose_format % (
                            item.get("creator") or "",
                            ct,
                            file_count, child_count, subset_count,
                            namespace + ":" + name
                        ))
                    else:
                        print(verbose_format % (
                            item.get("creator") or "",
                            ct,
                            namespace + ":" + name
                        ))
                else:
                    print("%s:%s" % (namespace, name))
                    

class ShowDatasetCommand(CLICommand):
    
    Opts = ("pj", ["pprint=","json"])
    Usage = """[<options>] <namespace>:<name>                           -- print dataset info
            -j|--json       - print as JSON
            -p|--pprint     - Python pprint
    """
    MinArgs = 1

    def __call__(self, command, client, opts, args):
        info = client.get_dataset(args[0])
        if "-p" in opts or "--pprint" in opts:
            pprint.pprint(info)
        elif "-j" in opts or "--json" in opts:
            print(json.dumps(info, indent=4, sort_keys=True))
        else:
            print("Namespace:       ", info["namespace"])
            print("Name:            ", info["name"])
            print("Description:     ", info.get("description", ""))
            print("Creator:         ", info.get("creator", ""))
            ct = info.get("created_timestamp") or ""
            if ct:
                ct = datetime.fromtimestamp(ct, timezone.utc).strftime("%Y-%m-%d %H:%M:%S %Z")
            print("Created at:      ", ct)
            print("Frozen:          ", "yes" if info.get("frozen", False) else "no")
            print("Monotonic:       ", "yes" if info.get("monotonic", False) else "no")
            print("Metadata:")
            if info.get("metadata"):
                print(indent(json.dumps(info["metadata"], indent=4, sort_keys=True), "  "))
            print("Constraints:")
            for name, constraint in sorted(info.get("file_meta_requirements", {}).items()):
                line = "  %-40s %10s" % (name, "required" if constraint.get("required", False) else "")
                if "values" in constraint:
                    line += " %s" % (tuple(constraint["values"]),)
                rng = None
                if "min" in constraint:
                    rng = [repr(constraint["min"]), ""]
                if "max" in constraint:
                    if rng is None: rng = ["", ""]
                    rng[1] = repr(constraint["max"])
                if rng is not None:
                    line += " [%s - %s]" % tuple(rng)
                if "pattern" in constraint:
                    line += " ~ '%s'" % (constraint["pattern"])
                print(line)
                    

class AddSubsetCommand(CLICommand):

    Usage = """<parent dataset namespace>:<parent name> <child dataset namespace>:<child name> [<child dataset namespace>:<child name> ...]
    """
    MinArgs = 2

    def __call__(self, command, client, opts, args):
        parent, children = args[0], args[1:]
        for child in children:
            client.add_child_dataset(parent, child)

def load_metadata(opts):
    # return None if no -j in options
    meta = None
    if "-m" in opts or "--metadata" in opts:
        arg = opts.get("-m") or opts.get("--metadata")
        if arg.startswith('@'):
            meta = json.load(open(arg[1:], "r"))
        else:
            meta = json.loads(arg)
    return meta
    
class CreateDatasetCommand(CLICommand):

    Opts = ("MFm:f:d:r:", ["monotonic", "frozen", "metadata=", "dataset-query=", "file-query=", "meta-requirements="])
    Usage = """[<options>] <namespace>:<name> [<description>]           -- create dataset
        -M|--monotonic
        -F|--frozen
        -m|--metadata '<JSON expression>'
        -m|--metadata @<JSON file>
        -f|--file-query '<MQL file query>'          - run the query and add files to the dataset
        -f|--file-query @<file_with_query>          - run the query and add files to the dataset
        -r|--meta-requirements '<JSON expression>'  - add metadata requirements
        -r|--meta-requirements @<JSON file>         - add metadata requirements
        """
    MinArgs = 1
    
    def get_text(self, opts, opt_keys):
        opt_value = None
        for key in opt_keys:
            opt_value = opts.get(key)
            if opt_value:   break
        if opt_value:
            if opt_value.startswith('@'):
                opt_value = open(opt_value[1:], "r").read()
        return opt_value

    def __call__(self, command, client, opts, args):
        dataset_spec, desc = args[0], args[1:]
        if desc:
            desc = " ".join(desc)
        else:
            desc = ""    
        monotonic = "-M" in opts or "--monotonic" in opts
        frozen = "-F" in opts or "--frozen" in opts
        metadata = load_metadata(opts) or {}
        files_query = self.get_text(opts, ["-f", "--file-query"])
        try:
            out = client.create_dataset(dataset_spec, monotonic = monotonic, frozen = frozen, description=desc, metadata = metadata,
                files_query = files_query
            )
        except MCError as e:
            print(e)
            sys.exit(1)
        else:
            print(out)

class UpdateDatasetCommand(CLICommand):

    Opts = ("MFm:r", ["replace", "monotonic", "frozen", "metadata="])
    Usage = """<options> <namespace>:<name> [<description>]             -- modify dataset info
            -M|--monotonic (yes|no) - set/reset monotonic flag
            -F|--frozen (yes|no)    - set/reset monotonic flag
            -r|--replace            - replace metadata, otherwise update
            -m|--metadata @<JSON file with metadata> 
            -m|--metadata '<JSON expression>' 
    """
    MinArgs = 1

    def __call__(self, command, client, opts, args):
        mode = "replace" if ("-r" in opts or "--replace" in opts) else "update"

        if not args or args[0] == "help":
            print(Usage)
            sys.exit(2)
        
        metadata = load_metadata(opts)

        dataset = args[0]
        monotonic = frozen = None
        if "-M" in opts or "--monotonic" in opts:    
            monotonic = opts.get("-M") or opts.get("--monotonic")
            if not monotonic in ("yes", "no"):
                print("Invalid value for -M or --monotonic option:", monotonic, ". Valid values are 'yes' and 'no'")
                sys.exit(2)
            monotonic = monotonic == "yes"
        if "-F" in opts or "--frozen" in opts:    
            frozen = opts.get("-F") or opts.get("--frozen")
            if not frozen in ("yes", "no"):
                print("Invalid value for -F or --frozen option:", frozen, ". Valid values are 'yes' and 'no'")
                sys.exit(2)
            frozen = frozen == "yes"
        desc = None
        if args[1:]:
            desc = " ".join(args[1:])

        try:
            response = client.update_dataset(dataset, metadata=metadata, frozen=frozen, monotonic=monotonic, mode=mode, description=desc)
        except MCError as e:
            print(e)
            sys.exit(1)
        else:
            print(response)

class AddFilesCommand(CLICommand):
    
    Opts = ("i:j:d:N:sq:", ["namespace=", "json=", "dids=", "ids=", "sample", "query="])
    Usage = """[options] <dataset namespace>:<dataset name>

            add files by DIDs or namespace/names
            -N|--namespace <default namespace>           - default namespace for files
            -d|--names <file namespace>:<file name>[,...]
            -d|--names -            - read the list from stdin
            -d|--names @<file>      - read the list from file

            add files by file id
            -i|--ids <file id>[,...]
            -i|--ids -              - read the list from stdin
            -i|--ids @<file>        - read the list from file

            add file list from JSON file
            -j|--json <json file>
            -j|--json -             - read JSON file list from stdin
            -s|--sample             - print JOSN file list sample

            add files matching a query
            -q|--query "<MQL query>"
            -q|--query @<file>      - read query from the file
    """
    
    AddSample = json.dumps(
        [
            {        
                "did":"test:file1.dat"
            },
            {        
                "namespace":"test",
                "name":"file2.dat"
            },
            {        
                "fid":"54634"
            }
        ],
        indent=4, sort_keys=True
    )

    def get_text(self, opts, opt_keys):
        opt_value = None
        for key in opt_keys:
            opt_value = opts.get(key)
            if opt_value:   break
        if opt_value:
            if opt_value.startswith('@'):
                opt_value = open(opt_value[1:], "r").read()
            if opt_value == "-":
                opt_value = sys.stdin.read()
        return opt_value

    def get_list(self, opts, opt_keys):
        opt_value = self.get_text(opts, opt_keys)
        out = []
        if opt_value:
            for line in opt_value.split():
                for word in line.split(","):
                    word = word.strip()
                    if word:
                        out.append(word)
        return out

    def __call__(self, command, client, opts, args):

        if "--sample" in opts or "-s" in opts:
            print(self.AddSample)
            sys.exit(0)

        if len(args) != 1:
            raise InvalidArguments("Invalid arguments")

        default_namespace = opts.get("-N")
        files = query = None
        if "-i" in opts or "--ids" in opts:
            files = [{"fid": fid} for fid in self.get_list(opts, ["-i", "--ids"])]
        elif "-d" in opts or "--dids" in opts:
            files = [{"did": did} for did in self.get_list(opts, ["-d", "--dids"])]
        elif "-j" in opts or "--json" in opts:
            json_file = opts.get("-j") or opts.get("--json")
            files = json.load(open(json_file, "r"))
        else:
            query = opts.get("-q") or opts.get("--query")
            
        if (query is None) == (files is None):
            raise InvalidArguments("Eitther file list or a query must be specified, but not both")

        dataset = args[-1]
        try:
            out = client.add_files(dataset, file_list=files, query=query)
        except MCError as e:
            print(e)
            sys.exit(1)
        else:
            print("Added", len(out), "files")

DatasetCLI = CLI(
    "create",       CreateDatasetCommand(),
    "show",         ShowDatasetCommand(),
    "files",        ListDatasetFilesCommand(),
    "list",         ListDatasetsCommand(),
    "add-subset",   AddSubsetCommand(),
    "add-files",    AddFilesCommand(),
    "update",       UpdateDatasetCommand()
)
    
 



Common = """

?query:  top_file_query             
    | top_dataset_query                            

qualified_name:     (FNAME ":")? FNAME

qualified_name_list:   qualified_name ("," qualified_name)*

?did:    FNAME ":" FNAME

fid_list:  FID ("," FID)*

//?did_list:  did ("," did)*

param_def_list :  param_def ("," param_def)*

param_def: FNAME "=" constant

?meta_exp:   meta_or                                                           

meta_or:    meta_and ( "or" meta_and )*

meta_and:   term_meta ( "and" term_meta )*

?term_meta:  scalar CMPOP constant                  -> cmp_op
    | scalar "in" constant ":" constant             -> in_range
    | scalar "not" "in" constant ":" constant       -> not_in_range
    | scalar "in" "(" constant_list ")"             -> in_set
    | scalar "not" "in" "(" constant_list ")"       -> not_in_set
    | META_NAME "present"                               -> present                   
    | META_NAME "not" "present"                         -> not_present                   
    | constant "in" META_NAME                           -> constant_in
    | constant "not" "in" META_NAME                     -> constant_not_in
    | "(" meta_exp ")"                              
    | "!" term_meta                                 -> meta_not
    | "exists" STRING                               -> json_path

scalar: META_NAME                                       -> meta_attribute
        | FILE_ATTR_NAME                                -> object_attribute
        | DATASET_ATTR_NAME                             -> object_attribute
        | META_NAME "[" "all" "]"                       -> array_all
        | META_NAME "[" "any" "]"                       -> array_any
        | META_NAME "[" SIGNED_INT "]"                  -> array_subscript
        | META_NAME "[" STRING "]"                      -> array_subscript
        | "len" "(" META_NAME ")"                       -> array_length

constant_list:    constant ("," constant)*                    

constant : SIGNED_FLOAT                             -> float_constant                      
    | STRING                                        -> string_constant
    | SIGNED_INT                                    -> int_constant
    | BOOL                                          -> bool_constant
    | UNQUOTED_STRING                               -> string_constant
    | "datetime" "(" (STRING|UNQUOTED_STRING) ")"   -> datetime_constant
    | "date" "(" (STRING|UNQUOTED_STRING) ( "," (STRING|UNQUOTED_STRING) )? ")"         -> date_constant

index:  STRING
    | SIGNED_INT

META_NAME: WORD ("." WORD)+                            // meta attribute has to have a dot in the name

FNAME: LETTER ("_"|"-"|"."|LETTER|DIGIT|"/")*

FID: ("_"|"-"|"."|LETTER|DIGIT|"/")+

WORD: LETTER ("_"|LETTER|DIGIT)*

CMPOP:  "<" "="? | "!"? "=" "="? | "!"? "~" "*"? | ">" "="? | "like"            //# like is not implemented yet

BOOL: "true"i | "false"i

STRING : /("(?!"").*?(?<!\\\\)(\\\\\\\\)*?"|'(?!'').*?(?<!\\\\)(\\\\\\\\)*?')/i
UNQUOTED_STRING : /[a-z0-9:%$@_^.%*?-]+/i

%import common.CNAME
%import common.SIGNED_INT
%import common.SIGNED_FLOAT

%import common.WS
%import common.LETTER
%import common.DIGIT
%ignore WS


"""


